//
//  ViewController.swift
//  BLE UT333 BT
//
//  Created by John Geelen on 14/02/2020.
//  Copyright © 2020 JG Electronics. All rights reserved.
//

import Cocoa
import CoreBluetooth

var txCharacteristic : CBCharacteristic?
var rxCharacteristic : CBCharacteristic?
var blePeripheral : CBPeripheral?
var characteristicASCIIValue = NSString()

class ViewController: NSViewController, CBCentralManagerDelegate, CBPeripheralDelegate  {

	@IBOutlet weak var tableView: NSTableView!
	@IBOutlet weak var temperatureTxtFld: NSTextField!
	@IBOutlet weak var moistureTxtFld: NSTextField!
	@IBOutlet weak var scanButton: NSButton!
	
	var centralManager : CBCentralManager!
	var peripherals: [CBPeripheral]? = []
	var RSSIs = [NSNumber]()
	var data = NSMutableData()
	var writeData: String = ""
	
	var characteristicValue = [CBUUID: NSData]()
	var characteristics = [String : CBCharacteristic]()
	
	var peripheralNameStr:String = ""
	var peripheralIdStr:String = ""
	var peripheralRssiStr:String = ""
	
	var timer = Timer()
	var counter: Bool = false
	
	var userDefaults = UserDefaults.standard
	
	
	
	override func viewDidLoad() {
		super.viewDidLoad()
		self.tableView.delegate = self
		self.tableView.dataSource = self
		centralManager = CBCentralManager(delegate: self, queue: nil)
		
	//  set timer
		let timer2 = Timer(timeInterval:2.0, target:self, selector:#selector(ViewController.fTimerFireMethod(_:)), userInfo:nil, repeats:true)
		let loop:RunLoop = RunLoop.current
		loop.add(timer2, forMode:RunLoop.Mode(rawValue: "NSDefaultRunLoopMode"))
	}
	
	override func viewDidAppear() {
	//	disconnectFromDevice()
	//	refreshScanView()
		print("View Cleared")
	}
	
	override func viewWillDisappear() {
		print("Stop Scanning")
		centralManager?.stopScan()
	}

	override var representedObject: Any? {
		didSet {
		// Update the view, if already loaded.
		}
	}
//	*********************************************************************************************************************************
	@objc func fTimerFireMethod(_ timer: Timer) {
	//	print("timer fired")
	//	print("counter = \(counter)")
		if counter == true {
		var ui8UdpBuffer = [UInt8] (repeating : 0x5E, count: 1)
		let data = (NSData(bytes:&ui8UdpBuffer, length:1)) as Data
		print("data = \(data[0])")
		print("sending data")
			blePeripheral!.writeValue(data, for: txCharacteristic!, type: CBCharacteristicWriteType.withoutResponse)
		}
	}
	
	@IBAction func refreshAction(_ sender: NSButton) {
		if(sender.state == NSControl.StateValue.on) {
			disconnectFromDevice()
			peripherals = []
			RSSIs = []
			tableView.reloadData()
			startScan()
		} else {
			cancelScan()
		}
	}
	
/*	Okay, now that we have our CBCentalManager up and running, it's time to start searching for devices. You can do this by calling the "scanForPeripherals" method.*/
		
	func startScan() {
		tableView.reloadData()
		peripherals = []
		print("Now Scanning...")
		self.timer.invalidate()
	//	let kBLEService_UUID = "19B10000-E8F2-537E-4F6C-D104768A1214"
	//	let BLEService_UUID = CBUUID(string: kBLEService_UUID)
	//	open func scanForPeripherals(withServices serviceUUIDs: [CBUUID]?, options: [String : Any]? = nil)
		centralManager?.scanForPeripherals(withServices:nil , options: [CBCentralManagerScanOptionAllowDuplicatesKey:false])
	//	centralManager?.scanForPeripherals(withServices:nil, options: [CBCentralManagerScanOptionAllowDuplicatesKey:false])
	//	Timer.scheduledTimer(timeInterval: 17, target: self, selector: #selector(self.cancelScan), userInfo: nil, repeats: false)
	}
		
/*	We also need to stop scanning at some point so we'll also create a function that calls "stopScan"*/
	func cancelScan() {
		self.centralManager?.stopScan()
		print("Scan Stopped")
		print("Number of Peripherals Found: \(String(describing: peripherals?.count))")
	//	peripheralsFound.stringValue = "\(String(describing: peripherals?.count))"
	}

	func refreshScanView() {
		tableView.reloadData()
	}
		
//	Peripheral Connections: Connecting, Connected, Disconnected
		   
//	Connection
	func connectToDevice () {
		centralManager?.connect(blePeripheral!, options: nil)
	}
		
//	Terminate all Peripheral Connection
/*
	Call this when things either go wrong, or you're done with the connection.
	This cancels any subscriptions if there are any, or straight disconnects if not.
	(didUpdateNotificationStateForCharacteristic will cancel the connection if a subscription is involved)
*/
	func disconnectFromDevice () {
		if blePeripheral != nil {
		//	We have a connection to the device but we are not subscribed to the Transfer Characteristic for some reason.
		// 	Therefore, we will just disconnect from the peripheral
			centralManager?.cancelPeripheralConnection(blePeripheral!)
		}
	}


//	*********************************************************************************************************************************
//	Central Manager Delegates
//	*********************************************************************************************************************************
	func restoreCentralManager() {
//	Restores Central Manager delegate if something went wrong
		centralManager?.delegate = self
	}

//	Called when the central manager discovers a peripheral while scanning. Also, once peripheral is connected, cancel scanning.
	func centralManager(_ central: CBCentralManager, didDiscover peripheral: CBPeripheral,advertisementData: [String : Any], rssi RSSI: NSNumber) {
//	func centralManager(_ central: CBCentralManager, didDiscover peripheral: CBPeripheral,advertisementData: [Dictionary Key: Any], rssi RSSI: NSNumber) {  key = CBAdvertisementDataManufacturerDataKey
		blePeripheral = peripheral
		print("Found peripheral")
	//	let number: NSNumber? = RSSI
		let peripheralRssiInt = RSSI.intValue
		print("peripheralRssiInt", peripheralRssiInt)
		if(!(peripherals?.contains(peripheral))!) {
			if(peripheral.name != nil) {
				if peripheralRssiInt > -100 {
					peripherals?.append(peripheral)
					RSSIs.append(RSSI)
					peripheral.delegate = self
					tableView.reloadData()
				}
			}
		}
		if blePeripheral != nil {
			print("")
			print("*************************************************************")
			print("Found new pheripheral devices with services")
			print("Peripheral name: \(String(describing: peripheral.name))")
			print ("Advertisement Data : \(advertisementData)")
			print ("RSSI: \(String(describing: RSSIs))")
			print("Peripheral info: \(String(describing: peripheral))")
			print("*************************************************************")
			print("")
		}
	}
		
	func centralManagerDidUpdateState(_ central: CBCentralManager) {
		print("centralManagerDidUpdateState")
		if central.state == CBManagerState.poweredOn {
		//	We will just handle it the easy way here: if Bluetooth is on, proceed...start scan!
			print("Bluetooth Enabled")
		//	startScan()
		} else {
		//	If Bluetooth is off, display a UI alert message saying "Bluetooth is not enable" and "Make sure that your bluetooth is turned on"
			print("Bluetooth Disabled- Make sure your Bluetooth is turned on")
		/*
		//	let alertVC = UIAlertController(title: "Bluetooth is not enabled", message: "Make sure that your bluetooth is turned on", preferredStyle: UIAlertControllerStyle.alert)
		//	let action = UIAlertAction(title: "ok", style: UIAlertActionStyle.default, handler: { (action: UIAlertAction) -> Void in self.dismiss(animated: true, completion: nil) }) alertVC.addAction(action)
			self.present(alertVC, animated: true, completion: nil)
		*/
		}
	}
		
/*
	Invoked when a connection is successfully created with a peripheral.
	This method is invoked when a call to connect(_:options:) is successful. You typically implement this method to set the peripheral’s delegate and to discover its services.
*/
//	Connected
	func centralManager(_ central: CBCentralManager, didConnect peripheral: CBPeripheral) {
		print("*****************************")
		print("Connection complete")
		print("Peripheral info: \(String(describing: blePeripheral))")
			   
	//	Stop Scan- We don't need to scan once we've connected to a peripheral. We got what we came for.
		centralManager?.stopScan()
		print("Scan Stopped")
			   
	//	Erase data that we might have
		data.length = 0
			   
	//	Discovery callback
		peripheral.delegate = self
	//	Only look for all services 	//	that matches transmit uuid replace nil met UUID
		peripheral.discoverServices(nil)
			   
	//	volgende in Basic Chat
	//	Once connected, move to new view controller to manager incoming and outgoing data
/*		let storyboard = UIStoryboard(name: "Main", bundle: nil)
		let uartViewController = storyboard.instantiateViewController(withIdentifier: "UartModuleViewController") as! UartModuleViewController
		uartViewController.peripheral = peripheral
		navigationController?.pushViewController(uartViewController, animated: true)
*/
	}
		
//	Invoked when the central manager fails to create a connection with a peripheral.
	func centralManager(_ central: CBCentralManager, didFailToConnect peripheral: CBPeripheral, error: Error?) {
		if error != nil {
			print("Failed to connect to peripheral")
			return
		}
	}
		
	func disconnectAllConnection() {
		centralManager.cancelPeripheralConnection(blePeripheral!)
	}
		
/*	Invoked when you discover the peripheral’s available services.
	This method is invoked when your app calls the discoverServices(_:) method.
	If the services of the peripheral are successfully discovered, you can access them through the peripheral’s services property.
	If successful, the error parameter is nil. If unsuccessful, the error parameter returns the cause of the failure.
*/
	func peripheral(_ peripheral: CBPeripheral, didDiscoverServices error: Error?) {
		print("*******************************************************")
		if ((error) != nil) {
			print("Error discovering services: \(error!.localizedDescription)")
			return
		}
		guard let services = peripheral.services else { return }
			
		print("Found \(services.count) services!")
	//	We need to discover the all characteristic
		for service in services {
			peripheral.discoverCharacteristics(nil, for:service)
			print("service UUID: \(service.uuid)")
		//	bleService = service
		}
		print("Discovered Services: \(services)")
	}
			   
/*	Invoked when you discover the characteristics of a specified service.
	This method is invoked when your app calls the discoverCharacteristics(_:for:) method.
	If the characteristics of the specified service are successfully discovered, you can access them through the service's characteristics property.
	If successful, the error parameter is nil. If unsuccessful, the error parameter returns the cause of the failure.
*/
	func peripheral(_ peripheral: CBPeripheral, didDiscoverCharacteristicsFor service: CBService, error: Error?) {
		print("*******************************************************")
		if ((error) != nil) {
			print("Error discovering services: \(error!.localizedDescription)")
		return
		}
		guard let characteristics = service.characteristics else { return }
		print("Found \(characteristics.count) characteristics!")
			
		for characteristic in characteristics {
			print(characteristic)
		//	looks for the right characteristic
		//	characteristic.uuid = BLE_Characteristic_uuid_FF12
		//	if characteristic.uuid.isEqual(BLE_Characteristic_uuid_FF12)  {
			//	peripheral.setNotifyValue(true, for: characteristic)
		//		print("BLE_Characteristic_uuid_FF12 found")
		//	}
			if characteristic.uuid.isEqual(BLE_Characteristic_uuid_FF02)  {
				peripheral.setNotifyValue(true, for: characteristic)
				print("BLE_Characteristic_uuid_FF02 found")
			//	peripheral.readValue(for: characteristic)		// kan niet moet eerst een notify krijgen maar die komt niet
			}
		/*
			if characteristic.uuid.isEqual(BLE_Characteristic_uuid_FF08)  {
				peripheral.setNotifyValue(true, for: characteristic)
				print("BLE_Characteristic_uuid_FF08 found")
				peripheral.readValue(for: characteristic)		// kan niet moet eerst een notify krijgen maar die komt niet
			}
		*/
			print("count: \(characteristics.count)")
			print("characteristic UUID: \(characteristic.uuid)")
			if characteristic.properties.contains(.read) {
				print("\(characteristic.uuid): properties contains .read")
				peripheral.readValue(for:characteristic)
				print("")
			}
			if characteristic.properties.contains(.notify) {
				print("\(characteristic.uuid): properties contains .notify")
				peripheral.setNotifyValue(true, for: characteristic)
			}

			if(characteristic.uuid.uuidString == "FF01")
			{
				print("characteristic FF01 write: \(characteristic.uuid)")
				txCharacteristic = characteristic
				blePeripheral = peripheral
			//	var buffer = [UInt8] (repeating: 0, count: 1)
			//	buffer[0] = 0x5E
			//	let data = NSData(UInt8:&buffer, lenght:1)
				var ui8UdpBuffer = [UInt8] (repeating : 0x5E, count: 1)
				let data = (NSData(bytes:&ui8UdpBuffer, length:1)) as Data
				print("sending data")
				print("data = \(data[0])")
				print(data)
			//	print(data as NSData)
				blePeripheral!.writeValue(data, for: txCharacteristic!, type: CBCharacteristicWriteType.withoutResponse)
				counter = true
			}
			
		}
	//	peripheral.discoverDescriptors(for:rxCharacteristic)
	}
			   
		
/*	Getting Values From Characteristic
	After you've found a characteristic of a service that you are interested in,
	you can read the characteristic's value by calling the peripheral "readValueForCharacteristic" method within the "didDiscoverCharacteristicsFor service" delegate.
*/
	func peripheral(_ peripheral: CBPeripheral, didUpdateValueFor characteristic: CBCharacteristic, error: Error?) {
		print("didUpdateValueFor characteristic UUID: \(characteristic.uuid)")
			
		let data = characteristic.value
		print("data update value \(String(describing: data))")
		if (data?.count == 19) {
			print("data = \(String(describing: data?[0]))")
			print("data = \(String(describing: data?[1]))")
			print("data = \(String(describing: data?[2]))")
			print("data = \(String(describing: data?[3]))")
			print("data = \(String(describing: data?[4]))")
			print("data = \(String(describing: data?[5]))")
			print("data = \(String(describing: data?[6]))")
			print("data = \(String(describing: data?[7]))")
			print("data = \(String(describing: data?[8]))")
			print("data = \(String(describing: data?[9]))")
			print("data = \(String(describing: data?[10]))")
			print("data = \(String(describing: data?[11]))")
			print("data = \(String(describing: data?[12]))")
			print("data = \(String(describing: data?[13]))")
			print("data = \(String(describing: data?[14]))")
			print("data = \(String(describing: data?[15]))")
			print("data = \(String(describing: data?[16]))")
			print("data = \(String(describing: data?[17]))")
			print("data = \(String(describing: data?[18]))")
			
			let buffer1: [UInt8] = [(data?[4])!]
			let nsdata1 = Data(bytes: buffer1 as [UInt8], count: buffer1.count)
			let myStr1 = String(data: nsdata1, encoding: .utf8)
			print(myStr1 as Any)
			
		//	temperatuur
			if(myStr1 == "0") {
				let buffer2: [UInt8] = [(data?[7])!,(data?[8])!,(data?[9])!, (data?[10])!]
				let nsdata2 = Data(bytes: buffer2 as [UInt8], count: buffer2.count)
				let myStr2 = String(data: nsdata2, encoding: .utf8)
				let tempStr2 = (myStr2 ?? "error") + " ˚C"
				print(myStr2 as Any)
				temperatureTxtFld.stringValue = tempStr2
			}
		//	moisture
			if(myStr1 == "4") {
				let buffer3: [UInt8] = [(data?[7])!,(data?[8])!,(data?[9])!, (data?[10])!]
				let nsdata3 = Data(bytes: buffer3 as [UInt8], count: buffer3.count)
				let myStr3 = String(data: nsdata3, encoding: .utf8)
				let tempStr3 = (myStr3 ?? "error") + " %RH"
				print(myStr3 as Any)
				moistureTxtFld.stringValue = tempStr3
			}
		}
	}
		
	func peripheral(_ peripheral: CBPeripheral, didUpdateNotificationStateFor characteristic: CBCharacteristic, error: Error?) {
		print("*******************************************************")
			
		if (error != nil) {
			print("Error changing notification state:\(String(describing: error?.localizedDescription))")
		} else {
			print("Characteristic's value subscribed")
		}
		if (characteristic.isNotifying) {
			print ("Subscribed. Notification has begun for: \(characteristic.uuid)")
		}
	}
}

//	following functions from Basic Chat
//	Write function for string
	func writeValue(data: String){
		let valueString = (data as NSString).data(using: String.Encoding.utf8.rawValue)
	//	change the "data" to valueString
		if let blePeripheral = blePeripheral{
			if let txCharacteristic = txCharacteristic {
				blePeripheral.writeValue(valueString!, for: txCharacteristic, type: CBCharacteristicWriteType.withResponse)
			}
		}
	}

// Write function for value
	func writeCharacteristic(val: Int8){
		var val = val
		let ns = NSData(bytes: &val, length: MemoryLayout<Int8>.size)
		blePeripheral!.writeValue(ns as Data, for: txCharacteristic!, type: CBCharacteristicWriteType.withResponse)
	}
	
//	*********************************************************************************************************************************
//	Table View Extension
//	*********************************************************************************************************************************
extension ViewController: NSTableViewDataSource, NSTableViewDelegate {
		
	func numberOfRows(in tableView: NSTableView) -> Int {
		print("myTableViewData.count", peripherals?.count as Any)
		return (peripherals?.count)!
	}
	
	func tableView(_ tableView: NSTableView, viewFor tableColumn: NSTableColumn?, row: Int) -> NSView?{
		let peripheral = self.peripherals![row]
		let RSSI = self.RSSIs[row]
		let RSSI_Str = String(describing: RSSI)
		
		let result:ScanTableCellView = tableView.makeView(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "defaultRow"), owner: self) as! ScanTableCellView
	//	result.imgView.image = NSImage(named:NSImage.Name(rawValue: "NSUser"))
			
		result.deviceNameTextField.stringValue = "no name"
			
		if(peripheral.name != nil) {
			result.deviceNameTextField.stringValue = peripheral.name!
		//	result.deviceID_TextField.stringValue = peripheral
			result.deviceRSSI_TextField.stringValue = RSSI_Str
		} else {
			result.deviceNameTextField.stringValue = "no name"
			print("no name")
		}
		return result;
	}
/*
	func tableViewSelectionDidChange(_ notification: Notification) {
		print("tableViewSelectionDidChange")

		if let myData = notification.object as? NSTableView {
	//	we create an [Int] array from the index set
			print(myData.selectedRow)
			if(myData.selectedRow >= 0) {
			
			let selected = myData.selectedRowIndexes.map {Int ($0)}
			print(selected[0])
				
			let deviceNameStr: String = myTableViewData[selected[0]]["deviceName"]!
			let deviceIdStr: String = myTableViewData[selected[0]]["deviceID"]!
			let deviceRssiStr: String = myTableViewData[selected[0]]["deviceRSSI"]!
				
			userDefaults.set(peripheralNameStr, forKey:"scanDeviceNameStr")
			userDefaults.set(peripheralIdStr, forKey:"scanDeviceIdStr")
			userDefaults.set(peripheralRssiStr, forKey:"scanDeviceRssiStr")
			userDefaults.synchronize()
					
			print("deviceName_2", deviceNameStr)
			print("deviceID_2", deviceIdStr)
			print("deviceRSSI_2", deviceRssiStr)
			}
		}
	}
*/
/*	func tableView(_ tableView: NSTableView, didSelectRowAt indexPath: IndexPath) {
		blePeripheral = peripherals?[indexPath.row]
		connectToDevice()
	}
*/
	func tableViewSelectionDidChange(_ notification: Notification) {
		print("tableViewSelectionDidChange")
		
		if let myData = notification.object as? NSTableView {
			print(myData.selectedRow)
			if(myData.selectedRow >= 0) {
		//	we create an [Int] array from the index set maar myData.selectedRow geeft al de geselecteerde rij dus waarom op onderstaande manier?
				let selected = myData.selectedRowIndexes.map {Int ($0)}
				print(selected[0])
				blePeripheral = peripherals?[selected[0]]
				let RSSI = self.RSSIs[selected[0]]
				peripheralRssiStr = String(describing: RSSI)
			//	peripheralNameStr = (blePeripheral?.name!)!
				peripheralIdStr = "???"
			/*
				userDefaults.set(peripheralNameStr, forKey:"scanDeviceNameStr")
				userDefaults.set(peripheralIdStr, forKey:"scanDeviceIdStr")
				userDefaults.set(peripheralRssiStr, forKey:"scanDeviceRssiStr")
				userDefaults.synchronize()
			*/
			
				print("peripheralName", peripheralNameStr)
				print("peripheralID", peripheralIdStr)
				print("peripheralRSSI", peripheralRssiStr)
			
				cancelScan()
				scanButton.state = NSControl.StateValue.off
				connectToDevice()
			}
		}
	}
}

